/*
    Copyright (c) 2002 Jens Tinz
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
    3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/**
 * @fileoverview This module integrates the radial menu with the browser.
 *
 * There's functionality for responding to events and reading
 * preferences. Entry points are {@link integrateOnLoad} and possibly
 * {@link Integrator} directly.
 */


// The code dealing with the normal context menu is fragile, especially
// in dragonly mode. Some background:
//
// The order of events on Gtk+/GNU/Linux in Firefox 3.0b5 seems to be:
// - mousedown happens (capture and bubble through the browser element).
// - contextmenu happens (capture through the browser element).
//   This seems to happen no matter what we do to the mousedown.
// - I *think* when the contextmenu bubble reaches the browser it shows the
//   popup and stops propagation (nsXULPopupListener).
//   This triggers popupshowing on the actual contextmenu (capture+bubble).
//   When agressive is true Mozilla ignores preventDefault on the contextmenu
//   event. However it does *not* ignore stopPropagation! Mozilla bug?
//   I think it should be doing the same thing I do and listen in the
//   capture phase when agressive...
//
// But on Windows the context menu is triggered on mouse *up*, not down.
// That is actually the easy case: we just go back to idle on mouseup
// and then let Mozilla handle the event normally. If the contextmenu
// gets triggered on mousedown we block it and remember that, then
// force it to appear on mouseup. Doing it this way should be more
// reliable than guessing if we should open it on mouseup (depending
// on mouse button and platform and whatnot) but the interaction
// with "agressive" mode is a bit fragile (depending on how Mozilla and
// we handle that we may block the context menu without realising it).
// For example: if blocking (stopPropagation/preventDefautl) the mouse
// event prevents the contextmenu event on some platform/Mozilla version
// we would have to modify our code to keep getting the context menu.
//
// We may have to do a similar "remember we prevented it on mousedown,
// retrigger on mouseup" dance for other things, but I have not run
// into them yet.
//
// This does not interact all that well with scripts calling alert()
// in response to mouse and/or contextmenu events. I may want to try
// to detect the focus loss to the browser or something like that.


const menu = Components.utils.import(
    'resource://radialcontext-mz/menu.js', null);
const actions = Components.utils.import(
    'resource://radialcontext-mz/actions.js', null);
const utils = Components.utils.import(
    'resource://radialcontext-mz/utils.js', null);

const nsIDOMHTMLDocument = Components.interfaces.nsIDOMHTMLDocument;
const nsIDOMXULDocument = Components.interfaces.nsIDOMXULDocument;
const nsIDOMXMLDocument = Components.interfaces.nsIDOMXMLDocument;
const XHTMLNS = 'http://www.w3.org/1999/xhtml';

// Constants: valid "enum" values for this.state.

// No radial menu triggered:
const kStateHidden = 0;
// Menu still invisible but mouse down (in "open on drag" mode):
const kStateOpenOnDrag = 1;
// Menu was just drawn, mouse button is down:
const kStateShowing = 2;
// Menu was dragged, mouse button is down:
const kStateDragged = 3;
// Mouse button was released without moving the mouse, menu stays
// open until any mouse button is clicked:
const kStateClickedOpen = 4;

let (prefService = Components.classes['@mozilla.org/preferences-service;1']
     .getService(Components.interfaces.nsIPrefService)) {
    var prefbranch = prefService.getBranch('extensions.radialcontext-mz.');
    prefbranch.QueryInterface(Components.interfaces.nsIPrefBranch2);
    var agressiveBranch = prefService.getBranch('dom.event.contextmenu.');
    agressiveBranch.QueryInterface(Components.interfaces.nsIPrefBranch2);
}

// Clear some prefs used by older versions.
// The tip dialog and click_for_normal were removed in 1.1.
// ttip_keep was moved into ttip_after_move in 1.1.
for each (let prefname in ['tips_at_startup', 'tip_nr', 'ttip_keep',
                           'click_for_normal']) {
    try {
        prefbranch.clearUserPref(prefname);
    } catch (e) {
        // Does not exist or has no user-set value. Ignore.
    }
}

// We use this a couple of times, so keep a reference (until shutdown):
var observerService = Components.classes['@mozilla.org/observer-service;1']
    .getService(Components.interfaces.nsIObserverService);

/**
 * @class Maybe display a firstrun page in the topmost window.
 * @implements nsITimerCallback
 * @implements nsIObserver
 */
function FirstrunHandler() {
    // Register us as observer for the "session restored" event if the
    // version of this extension currently running does not match the
    // one used on the previous run:
    this.installedVersion = Components.classes
        ['@mozilla.org/extensions/manager;1']
        .getService(Components.interfaces.nsIExtensionManager)
        .getItemForID('radialcontext-mz@mzz.mine.nu')
        .version;
    if (!prefbranch.prefHasUserValue('lastrun_version') ||
        prefbranch.getCharPref('lastrun_version') !== this.installedVersion)
        observerService.addObserver(this, 'sessionstore-windows-restored',
                                    false);
}

FirstrunHandler.prototype = {
    observe: function(subject, topic, data) {
        if (topic !== 'sessionstore-windows-restored')
            return;

        // HACK: If I just show the tab now it does not work reliably
        // (in Firefox 3.0b5). So use a timer. We have to keep the
        // timer alive until it fires, so we cannot call
        // removeObserver yet, since that is keeping us alive.
        this.timer = utils.Timer(this, 1000, utils.TIMER_ONE_SHOT);
    },

    notify: function(timer) {
        // Probably not necessary, but does not hurt.
        this.timer = null;
        observerService.removeObserver(this, 'sessionstore-windows-restored');

        // Based on code from fuelApplication.js.
        Components.classes
            ['@mozilla.org/appshell/window-mediator;1']
            .getService(Components.interfaces.nsIWindowMediator)
            .getMostRecentWindow('navigator:browser')
            .document.getElementById('cmd_radialcontext_mz_about')
            .doCommand();

        // Do this last so we try again next time if we failed to display now:
        prefbranch.setCharPref('lastrun_version', this.installedVersion);
    }
};

// The instance will be kept alive by the observer service if necessary.
new FirstrunHandler();


// prepare menu actions/layouts and register customization pref listener:
var actionManager = new actions.ActionManager(prefbranch);
// (This is actually slightly dodgy, since if you open two radial menus
// simultaneously in different browser windows they share their Action
// instances, so tweak() calls in one window affect the other window too.
// I am counting on nobody being crazy enough to click open simultaneous
// menus like that. If that assumption turns out to be wrong this should
// just go back to being an instance var of Integrator.)

// Shut down the actions instance when Mozilla shuts down.
// Is quit-application the right topic? Probably close enough.
let (shutdownObserver = {
        observe: function(subject, topic, data) {
            if (topic === 'quit-application') {
                actionManager.shutdown(prefbranch);
                observerService.removeObserver(this, 'shutdown');
                observerService = null;
            }
        }
     }) {
    observerService.addObserver(shutdownObserver, 'quit-application', false);
}


/**
 * @class Event handler for a window.
 * @implements nsIDOMEventListener
 * @implements nsITimerCallback
 * @implements nsIObserver
 * @param {nsIDOMWindow} window The browser window.
 *   Must support getBrowser() and getMarkupDocumentViewer() among others.
 */
function Integrator(window) {
    this.window = window;
    // Current state of the menu/mouse (see kState* for values):
    this.state = kStateHidden;
    // Menu currently displayed. Invariant: is null if and only if
    // state is kStateHidden.
    this.menu = null;
    // Time to wait after the menu moved before redisplaying labels,
    // or -1 to hide without redisplay, or 0 to leave them alone.
    this.tooltipDelay = -1;
    // Timeout for displaying the labels:
    this.labeltimer = utils.RawTimer();
    // Offset from screen to client coordinates
    this.screenToClientX = 0;
    this.screenToClientY = 0;
    // These two are used to work around two event handling glitches:
    // spurious mousemove events and omitted mousemove events.
    this.lastMouseX = null;
    this.lastMouseY = null;
    // To convert our screen coordinates to client coordinates.
    this.fullZoom = 1.0;
    // To show the menu on mouseup in kStateOpenOnDrag
    // after blocking it on mousedown:
    this.blockedNormalContextMenu = false;

    agressiveBranch.addObserver('enabled', this, false);
    // Whether we are preventing other scripts from interfering with the menu
    this.agressive = !agressiveBranch.getBoolPref('enabled');

    // Add some event listeners:
    this.addEventListeners();
    window.addEventListener('unload', this, false);
}

Integrator.prototype = {
    /**
     * Called when our window goes away.
     */
    on_unload: function(event) {
        // Drops the reference to the callback (us).
        this.labeltimer.cancel();
        agressiveBranch.removeObserver('enabled', this);
        this.removeEventListeners();
        this.window.removeEventListener('unload', this, false);
    },

    /**
     * Called by labeltimer to display the tooltip.
     */
    notify: function(timer) {
        this.menu.showLabels();
    },

    /**
     * Go to kStateHidden.
     * Can be called from any state (including kStateHidden).
     *
     * @param {bool} doAction If true also execute the selected action.
     */
    _hideMenu: function(doAction) {
        if (this.state === kStateHidden)
            return;

        // Do this first to avoid getting stuck if this function malfunctions:
        this.state = kStateHidden;
        // Calling this again for an already-unregistered listener is
        // probably safe.
        this.window.getBrowser().removeEventListener('mousemove', this, true);
        // Assume this is safe even if the timer is not running.
        this.labeltimer.cancel();
        // The actual action functions are probably the most
        // frequently broken code, so make sure we deal if this
        // raises:
        try {
            this.menu.close(doAction);
        } finally {
            let context = this.menu.context;
            this.menu = null;
            context.dehighlight();
        }
    },

    handleEvent: utils.handleEvent,

    browserEvents: [
        'mousedown', 'mouseup', 'keydown', 'keyup', 'contextmenu'],

    /**
     * Register event listeners on the browser and some other stuff.
     */
    addEventListeners: function() {
        let browser = this.window.getBrowser();
        for each (let eventType in this.browserEvents)
            browser.addEventListener(eventType, this, this.agressive);

        this.window.document.getElementById('contentAreaContextMenu')
            .addEventListener('popupshowing', this, this.agressive);
    },

    /**
     * Unregister everything addEventListeners registered.
     *
     * Do not change this.agressive between addEventListeners and
     * removeEventListeners.
     */
    removeEventListeners: function() {
        let browser = this.window.getBrowser();
        for each (let eventType in this.browserEvents)
            browser.removeEventListener(eventType, this, this.agressive);

        this.window.document.getElementById('contentAreaContextMenu')
            .removeEventListener('popupshowing', this, this.agressive);
    },

    observe: function(branch, topic, data) {
        if (topic !== 'nsPref:changed')
            throw new Error('huh, I did not sign up for ' + topic);

        if (data !== 'enabled')
            return;

        this.removeEventListeners();
        this.agressive = !agressiveBranch.getBoolPref('enabled');
        this.addEventListeners();
    },

    on_contextmenu: function(evt) {
        if (this.agressive && this.state !== kStateHidden) {
            // We are already displayed or waiting in open-on-drag,
            // and we do not want anything interfering.
            // Especially not silly content-level scripts calling alert()
            // oncontextmenu. So intercept this event:
            evt.preventDefault();
            evt.stopPropagation();
            // Depending on the "agressive" pref (and potentially the Mozilla
            // version) we may or may not have blocked the context menu
            // (and its popupshowing event), so we need to do this here,
            // not in popupshowing:
            this.blockedNormalContextMenu = true;
        }
        // afaict evt.defaultPrevented is the DOM3Event standard,
        // getPreventDefault is from nsIDOMNSUIEvent and has the advantage
        // of existing.
        if (!this.agressive && evt.getPreventDefault())
            // Someone is trying to prevent a context menu and we are
            // willing to cooperate.
            this._hideMenu(false);
    },

    /**
     * Maybe suppress the standard context menu.
     */
    on_popupshowing: function(evt) {
        // Showing the normal menu while the radial one is open is
        // always bad. We also block the menu while waiting for drag
        // in drag-on-open.
        //
        // Notice we still block the menu in kStateOpenOnDrag even if
        // the context menu is not on "our" mouse button, but I think
        // that is ok (it is definitely no big deal).
        if (this.state !== kStateHidden)
            evt.preventDefault();
            // No point in calling stopPropagation here, this does not
            // bubble through content.
    },

    /**
     * Change menu entries on alt, cancel menu on esc.
     */
    on_keydown: function(evt) {
        if (this.menu !== null) {
            switch (evt.which) {
            case evt.DOM_VK_ALT:
                this.menu.updateEntries(true);
                // On windows hitting alt activates the menu bar, so we
                // need to preventDefault here (TODO: actually test this
                // change on windows)
                evt.preventDefault();
                break;
            case evt.DOM_VK_ESCAPE:
                this._hideMenu(false);
                // not 100% sure preventDefault is sane here, but hey,
                // just hit esc again, we are no longer listening.
                evt.preventDefault();
                break;
            }
        }
    },

    /**
     * Change menu entries on alt.
     */
    on_keyup: function(evt) {
        if (this.menu !== null && evt.which === evt.DOM_VK_ALT) {
            this.menu.updateEntries(false);
            // See the DOM_VK_ALT case in onKeyDown
            evt.preventDefault();
        }
    },

    /**
     * Handle mouse button down: maybe display the menu.
     */
    on_mousedown: function(evt) {
        // Completely ignore this event if we are being nice and someone
        // is trying to cancel the event (this does not actually work
        // in the limited testing I did, but meh).
        if (!this.agressive && evt.getPreventDefault())
            return;

        // Ignore mousedown here (we activate events in mouseup instead):
        if (this.state === kStateClickedOpen) {
            // but we do need to eat the event (especially if agressive).
            evt.preventDefault();
            if (this.agressive)
                evt.stopPropagation();
            return;
        }

        if (evt.button !== prefbranch.getIntPref('mbutton')) {
            // Toggle alt-actions on alt-click.
            if (this.menu !== null) {
                this.menu.updateEntries(!this.menu.altPressed);
                // and eat the event (agressive, menu on left
                // mouse, page with alert on right mouse...)
                evt.preventDefault();
                if (this.agressive)
                    evt.stopPropagation();
            }
            return;
        }

        if (this.menu !== null) {
            // We got a mousedown for "our" button while the menu is
            // being dragged. That should be impossible, but we
            // sometimes miss mouseup events. To make it harder to
            // shoot yourself in the foot by randomly hitting buttons
            // we hide the menu without executing an action here.

            // Ideally we would also do this on mousedown for an alt
            // button, but that is impossible for the same reason
            // we cannot detect this in mousemove: no obvious way to
            // get the current mouse button state. No big deal though,
            // since the alt button cannot do anything destructive.
            this._hideMenu(false);
            return;
        }

        // make sure we got the right button and modifiers
        let mask = (evt.ctrlKey ? 1 : 0) | (evt.shiftKey ? 2 : 0);
        let showMask =
            (prefbranch.getBoolPref('show_ctrl') ? 1 : 0) |
            (prefbranch.getBoolPref('show_shift') ? 2 : 0);
        let supprMask =
            (prefbranch.getBoolPref('suppr_ctrl') ? 1 : 0) |
            (prefbranch.getBoolPref('suppr_shift') ? 2 : 0);
        if ((showMask & mask) !== showMask || (
            (supprMask & mask) === supprMask && supprMask !== 0))
            return;

        let targetDoc = evt.originalTarget.ownerDocument;
        // Bail if the browser chrome is targeted.
        if (targetDoc === this.window.document)
            return;

        // Check for a usable document type
        let parentNode;
        if (targetDoc instanceof nsIDOMHTMLDocument)
            parentNode = targetDoc.body;
        // Bah, this does not actually work. The menu gets added to the tree
        // just fine but is invisible.
        // else if (targetDoc instanceof nsIDOMSVGDocument)
        //     parentNode = targetDoc.documentElement;
        // Making this customizable might be overkill.
        // The rationale is that I like this when working on radialcontext
        // dialogs displayed in browser tabs but others might have
        // more complicated XUL content that needs the right mouse button.
        else if (targetDoc instanceof nsIDOMXULDocument &&
                 prefbranch.getBoolPref('show_on_xul'))
            parentNode = targetDoc.documentElement;
        else if (targetDoc instanceof nsIDOMXMLDocument) {
            // I see this for http://pear.math.pitt.edu/
            // mathzilla/Examples/markupOftheWeek.mhtml
            // which starts with
            // <?xml version="1.0"?>
            // <?xml-stylesheet href="mml.css" type="text/css"?>
            // <!DOCTYPE html SYSTEM "mathml.dtd">
            // <html xmlns="http://www.w3.org/1999/xhtml"
            // xmlns:math="http://www.w3.org/1998/Math/MathML"
            // xmlns:xlink="http://www.w3.org/1999/xlink">
            // There are probably cases where we cannot support this,
            // but if there is an xhtml html and body node we should
            // work just fine using that.

            // Note this case does not trigger for "normal" xhtml documents:
            // those are nsIDOMHTMLDocument which does not inherit from
            // nsIDOMXMLDocument.

            let documentElement = targetDoc.documentElement;
            if (documentElement.namespaceURI === XHTMLNS &&
                documentElement.tagName === 'html') {
                // Find a direct descendant that is a body node
                // (no getElementsByTagNameNS, just direct children):
                let bodies = [node
                              for each (node in documentElement.childNodes)
                              if (node.nodeType === node.ELEMENT_NODE &&
                                  node.namespaceURI === XHTMLNS &&
                                  node.tagName === 'body')];
                if (bodies.length === 1)
                    parentNode = bodies[0];
                else
                    return;
            }
        } else
            return;

        // Set up the menu context. Careful: we may still abort (if we
        // find we are in an input element and unwanted) so do not
        // touch "this" yet!

        let selection = evt.view.getSelection();

        // For highlighting the (i)frame we target but not the entire
        // document if there are no frames.
        let iframeDoc = null;
        if (this.window.content.document !== targetDoc)
            iframeDoc = targetDoc.body;

        let activeOutline = null;
        if (prefbranch.getBoolPref('highlight_enabled'))
            // the "solid" in this could easily be customizable too,
            // but that would bloat the prefpane even more than it
            // already is. Perhaps this should be a hidden pref
            // though?
            activeOutline = (
                prefbranch.getIntPref('highlight_size') + 'px solid ' +
                prefbranch.getCharPref('highlight_color'));

        // Build a linked list of contexts (one per interesting node)
        // that share common attributes (like the selection) through
        // their __proto__. context.next points to the next (less
        // specific) context. The final context.next is null.

        let globalContext = {
            window: this.window,
            frameDoc: targetDoc,
            selection: selection.toString()
        };
        let context = null;
        // XXX imho we should not need targetURI: links and images
        // should expose a getter for an nsIURI. But I do not think
        // they currently do that (see also bug 399490 comment #6).
        function chainContext(layout, target) {
            return {
                layout: layout,
                target: target,
                // Set only for nodes where it makes sense (a, image):
                targetURI: null,
                next: context,
                activeOutline: activeOutline,
                originalOutline: null,
                highlight: function() {
                    if (this.activeOutline !== null && this.target !== null) {
                        this.originalOutline = this.target.style.outline;
                        this.target.style.outline = this.activeOutline;
                    }
                },
                dehighlight: function() {
                    if (this.originalOutline !== null && this.target !== null)
                        this.target.style.outline = this.originalOutline;
                },
                __proto__: globalContext
            };
        }
        // Start building the context chain at the least specific context:
        context = chainContext('scroll', iframeDoc);

        let mouseInSelection = false;
        for (let i = 0; i < selection.rangeCount; ++i) {
            let range = selection.getRangeAt(i);
            if (range.isPointInRange(evt.rangeParent, evt.rangeOffset)) {
                mouseInSelection = true;
                break;
            }
        }

        if (context.selection && !mouseInSelection)
            // There is a selection but other context is more important,
            // so add this here at the end of the chain:
            context = chainContext('selection', null);

        context = chainContext('navigation', null);

        let ioService = Components.classes
            ['@mozilla.org/network/io-service;1']
            .getService(Components.interfaces.nsIIOService);

        // XXX this loop is backwards: it adds the deepest (most
        // specific) node at the end (least specific) of the chain,
        // but that is actually useful for the very common
        // image-in-link case (we are more interested in the link than
        // in the image). Would it be better to gather contexts in the
        // correct order and manually push some types to the
        // top/bottom? Are there actual cases other than image-in-link
        // we care about?
        for (let node = evt.originalTarget; node !== null;
             node = node.parentNode) {
            if (node.nodeType === node.ELEMENT_NODE && (
                    node.namespaceURI === null ||
                    node.namespaceURI === XHTMLNS)) {
                switch (node.localName.toLowerCase()) {
                case 'area':
                case 'a':
                    // Ignore links with no href.
                    if (node.href) {
                        let targetURI = ioService.newURI(node.href, null,
                                                         node.baseURIObject);
                        if (targetURI.schemeIs('mailto'))
                            context = chainContext('mail', node);
                        else
                            context = chainContext('link', node);
                        context.targetURI = targetURI;
                    }
                    break;
                case 'img':
                    context = chainContext('image', node);
                    context.targetURI = ioService.newURI(node.src, null,
                                                         node.baseURIObject);
                    break;
                case 'input':
                    // Don't hinder working with input fields:
                    if (evt.button === 0)
                        return;
                    let nodetype = node.type.toLowerCase();
                    if (nodetype !== 'text' && nodetype !== 'password')
                        break;
                    // Fall through, treat as a textarea.
                case 'textarea':
                    // Don't hinder working with input fields:
                    if (evt.button === 0)
                        return;
                    context = chainContext('edit', node);
                    break;
                }
            }
        }

        if (mouseInSelection)
            // Primary context is selection.
            context = chainContext('selection', null);

        // If we got this far we are really going to respond to this event.

        this.lastMouseX = evt.clientX;
        this.lastMouseY = evt.clientY;

        // So where do we listen for events?
        //
        // If we listen on the main content document we do not receive
        // events if the mouse is over an iframe *unless* a node
        // "belonging" to the main document is displayed on top of the
        // iframe. Our menu is such an element. This means the menu
        // gets "stuck" if you move the mouse *quickly* into the
        // iframe.
        //
        // If we listen on getBrowser() we keep getting the events,
        // but evt.client{X,Y} "jumps" when the mouse moves to a
        // different document. I cannot find an api to convert the
        // coordinates, so instead we use screen coordinates and just
        // remember the offset for this first event and use that to
        // correct future events.
        //
        // TODO: when the mousedown has an iframe as target it would be
        // even better to attach the menu to the parent document instead
        // of the iframe, but I do not know how to determine the screen ->
        // client coordinate offsets in that case.
        // (nsIDOMNSDocument.getBoxObjectFor(parentNode) could probably
        // work but in Firefox 3.0rc1 that triggers a deprecation warning.
        // It tells me to use getBoundingClientRect instead but that
        // has client coordinates, so it is useless here.)

        this.window.getBrowser().addEventListener('mousemove', this, true);

        // We stop the contextmenu in a different way because mozilla's
        // "do not allow js to mess with the contextmenu" pref makes this
        // not work. But we still need this when we are not using the
        // right mouse button.
        evt.preventDefault();
        if (this.agressive)
            evt.stopPropagation();

        // Dragonly is implemented by constructing an initially hidden menu.
        // That is a bit inefficient when we do not get a drag event,
        // but simplifies the implementation: we do not have to invent a way
        // to store the context we just built until the menu is dragged.
        let dragonly = prefbranch.getBoolPref('dragonly');
        // textZoom getting should be in menu.js, but I cannot find a
        // browser.js-independent way of doing it.
        //
        // getMarkupDocumentViewer is defined by browser.js as
        //   gBrowser.markupDocumentViewer
        // which is an xbl property on a "browser" object:
        //   this.docShell.contentViewer
        //       .QueryInterface(nsIMarkupDocumentViewer)
        // with docShell an xbl property:
        //   this.boxObject.QueryInterface(nsIBrowserBoxObject)
        //
        // nsIDOMWindow has a textZoom attribute, but it is marked
        // noscript (why not just "readonly" is beyond me).
        // nsIMarkupDocumentViewer and nsIDocShell have (text)zoom
        // attributes, but I see no clean way of getting them in the idl:
        // only nsIXULWindow, nsIBrowserBoxObject and nsIIFrameBoxObject
        // have docShell attrs. So I see no obvious clean way of getting
        // at the zoom value from just a document or window object, which
        // is why the menu cannot figure it out on its own.
        let markupDocumentViewer = this.window.getMarkupDocumentViewer();

        // Compensate for full page zoom. The menu uses client
        // (zoomed) coordinates. We use screen (unzoomed) coordinates
        // and convert them right before passing them down to the
        // menu.
        this.fullZoom = markupDocumentViewer.fullZoom;
        this.screenToClientX = evt.clientX - evt.screenX / this.fullZoom;
        this.screenToClientY = evt.clientY - evt.screenY / this.fullZoom;
        this.blockedNormalContextMenu = false;
        let theme = new menu.Theme();
        this.menu = new menu.PieMenu(
            theme, parentNode, actionManager.layouts[context.layout],
            evt.screenX / this.fullZoom + this.screenToClientX,
            evt.screenY / this.fullZoom + this.screenToClientY,
            evt.altKey, markupDocumentViewer.textZoom, !dragonly, context);

        // As far as this code is concerned ttip_enabled could be
        // merged with ttip_after_move as a single "tooltip mode"
        // pref, but the prefpane can be prettier if they are separate
        // (it shows a menulist for after_move and a separate global
        // on/off toggle, and wants to persist the menulist setting
        // even when the toggle is off). Similarly this code cannot
        // attach any special meaning to ttip_delay <= 0 (like -1
        // meaning disabled): the prefpane might want to show that as
        // a slider + some other control for the <= 0 values and that
        // would not work.
        if (prefbranch.getBoolPref('ttip_enabled')) {
            // Could specialcase delay 0 here, but no real need, I think.
            let tooltipDelay = prefbranch.getIntPref('ttip_delay');
            this.labeltimer.initWithCallback(
                this, tooltipDelay, this.labeltimer.TYPE_ONE_SHOT);

            // Could move this outside the ttip_enabled check for more
            // extra confusing tooltip modes...

            // This intentionally makes "keep if displayed" and
            // "reset delay" identical if the delay is 0. Hiding and
            // immediately showing the menu just causes useless flashing.
            switch (prefbranch.getIntPref('ttip_after_move')) {
            case 0:
                // normal delay mode:
                this.tooltipDelay = tooltipDelay;
                break;
            case 1:
                // keep if displayed mode:
                this.tooltipDelay = 0;
                break;
            case 2:
                // hide after move mode:
                this.tooltipDelay = -1;
                break;
            default:
                // Blowing up on an invalid pref setting would be lame.
                // Fall back to default and log the error instead.
                this.tooltipDelay = tooltipDelay.
                Components.utils.reportError(
                    'RadialContext: invalid value for ttip_after_move pref!');
                break;
            }
        } else
            // the mousemove code needs to do nothing for both keep if
            // displayed and completely disabled:
            this.tooltipDelay = 0;

        this.state = dragonly ? kStateOpenOnDrag : kStateShowing;

        // We do this even in dragonly mode, but I think that is ok.
        context.highlight();
    },

    /**
     * Deal with mousemove (pop up menu if necessary, move it).
     */
    on_mousemove: function(evt) {
        // The check for a coordinate change here is necessary because
        // of mozilla bug 321794: we get a spurious move event when
        // our menu div appears under the mouse cursor.
        if (this.lastMouseX === evt.clientX &&
            this.lastMouseY === evt.clientY)
            return;
        this.lastMouseX = evt.clientX;
        this.lastMouseY = evt.clientY;

        switch (this.state) {
        case kStateHidden:
            // Paranoia. Unregister so we do not flood the error message:
            this._hideMenu(false);
            throw new Error(
                'radialcontext bug: mousemove handler not unregistered');
        case kStateOpenOnDrag:
            this.menu.setVisible(true);
            // fall through
        case kStateShowing:
            this.state = kStateDragged;
            // fall through
        case kStateDragged:
            // It would be very useful if we could check the current mouse
            // button state here. The button *should* be down in this state,
            // but in some cases the mouse button comes up without us
            // receiving a mouseup event, causing the menu to get "stuck".
            // Unfortunately I cannot find a DOM method or anything
            // moz-specific in the idl files to check the current button
            // state...

            // fall through
        case kStateClickedOpen:
            if (this.menu.update(
                    evt.screenX / this.fullZoom + this.screenToClientX,
                    evt.screenY / this.fullZoom + this.screenToClientY)) {
                // Menu moved, muck with the tooltip.
                if (this.tooltipDelay < 0) {
                    this.labeltimer.cancel();
                    this.menu.hideLabels();
                } else if (this.tooltipDelay > 0) {
                    // We currently do not distinguish having a tooltip up
                    // (need to hide it and reinit the timer) and having
                    // one pending (need to reset the timer's delay).
                    // We simply always completely reinitialize.
                    this.menu.hideLabels();
                    this.labeltimer.initWithCallback(
                        this, this.tooltipDelay,
                        this.labeltimer.TYPE_ONE_SHOT);
                }
            } else if (this.tooltipDelay > 0)
                // If we have a tooltip pending, push it back.
                // I expect this to have no effect if the timer is cancelled.
                this.labeltimer.delay = this.tooltipDelay;
            break;
        default:
            throw new Error(
                'radialcontext bug: unhandled case in onMouseMove');
        }
    },

    /**
     * Deal with a mouse button being released.
     */
    on_mouseup: function(evt) {
        // We only care about other mouse buttons in clicked-open mode.
        // In the other modes it toggles alt state on mouse down, not up.
        // Corner case: we stay in kStateOpenOnDrag when other buttons
        // are clicked.
        if (this.state !== kStateClickedOpen &&
            evt.button !== prefbranch.getIntPref('mbutton'))
            return;

        switch (this.state) {
        case kStateOpenOnDrag:
            // We were not dragged. Stop paying attention.

            // XXX I think this could be wrong (and _withPopupNode in actions
            // too) if the primary context is something like the selection
            // (with a null target) but there is a context with a nontrivial
            // target after that. We should probably grab the topmost
            // nonnull target before falling back to frameDoc.body.
            let target =
                this.menu.context.target || this.menu.context.frameDoc.body;
            this._hideMenu(false);
            if (this.blockedNormalContextMenu) {
                evt.preventDefault();
                // This is *critical*. Make sure to test thoroughly before
                // removing this. If you get this wrong the contextmenu
                // tends to use the popupNode set by the last successfully
                // opened context menu. If no node is set it gets
                // spectacularly confused. Since various things
                // (like our own action code) could set popupNode too
                // make sure the context menu always works correctly
                // (agressive on/off, various platforms, etc) before
                // messing with this.
                this.window.document.popupNode = target;
                this.window.document.getElementById('contentAreaContextMenu')
                    .openPopupAtScreen(evt.screenX, evt.screenY, true);
            }
            // It is important not to call preventDefault here
            // unless we manually opened the context menu, so
            // things like clicking on links with our menu on the
            // left mouse button work (on some but not all
            // platforms this is how the normal context menu is opened)
            break;
        case kStateShowing:
            if (this.lastMouseX !== evt.clientX ||
                this.lastMouseY !== evt.clientY)
                // We did not get a mousemove event (or we would be in
                // kStateDragged) but the mouse did move. I think this
                // happens occasionally when the browser and/or system
                // is busy: you click and move the mouse, nothing
                // happens, you release the mouse, and after a while
                // the menu pops up in clicked-open mode. This is
                // annoying (you have to carefully click in the middle
                // to get rid of the menu) so close the menu instead.
                //
                // TODO It might make sense to check the time between
                // the mousedown and mouseup event: if they are close
                // together we can probably accept them (to allow
                // *really* fast back/forward gestures). But that
                // involves figuring out the right magic number for
                // max delay there, which is no fun.
                this._hideMenu(false);
            else {
                evt.preventDefault();
                this.state = kStateClickedOpen;
            }
            break;
        case kStateDragged:
        case kStateClickedOpen:
            evt.preventDefault();
            this._hideMenu(true);
            break;
        case kStateHidden:
            // We get here after the following sequence:
            // - Open the menu in kStateClickedOpen mode.
            // - Press and hold "our" mouse button and another mouse button.
            //   We ignore those events.
            // - Release the one that is not "our" button. We close the menu.
            // - Release "our" button.
            // In this case we want to prevent the popup menu from popping up.
            // (I think this is only really necessary on windows, where
            // the context menu is popped up on mouseup.)

            // Actually, we probably also get here after opening the
            // regular menu through ctrl+rightmouse. Is preventDefault
            // ok then? I think it is, and stuff seems to work, so
            // leave this alone for now.
            evt.preventDefault();
            break;
        default:
            throw new Error('radialcontext bug: unhandled case in onMouseUp');
        }
    }
};


function _integrate(event) {
    let window = event.target.defaultView;
    new Integrator(window);
    window.removeEventListener('load', _integrate, true);
}

/**
 * Helper to integrate with a window once it has fully loaded.
 */
function integrateOnLoad(window) {
    window.addEventListener('load', _integrate, true);
}
