/*
    Copyright (c) 2002 Jens Tinz
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
    3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/**
 * @fileoverview This module implements the actual radial menu.
 *
 * It can be attached to HTML documents and updated in response to
 * events. To use this you need to have a DOM document to attach to
 * and you need to write some glue to drive our event handlers.
 */


/**
 * @class Instances of this tell the PieMenu how it should look.
 *
 * This is still in flux.
 */

function Theme() {}

Theme.prototype = {
    // Path to icons base directory (trailing '/' is required).
    path: 'chrome://radialcontext-mz/skin/',
    // Coordinates of the menu center in the central "menu" image:
    menuXOffset: 46,
    menuYOffset: 46,
    // Radius of the hole (no selection until mouse is outside this):
    holeRadius: 17,
    // Distance from menu center to center of action icon:
    imagesRadius: 30,
    // Distance at which the menu starts to move or a submenu opens:
    actionRadius: 36,
    // Coordinates of the center of an icon:
    iconXOffset: 11,
    iconYOffset: 11,
    // Coordinates of the center of the "labels" image:
    labelsMainXOffset: 52,
    labelsMainYOffset: 55,
    // Distance from the top of the image to the top of the first label:
    labelsLeftYOffset: 1,
    labelsRightYOffset: 23,
    // Height of a single label:
    labelHeight: 22,
    // Font size used in labels. Usually a bit smaller than labelHeight.
    // (Remember this is in pixels, not points).
    fontSize: 12,

    // Map from input image name to actual image name.
    // I think this is a wart and using symlinks would be better,
    // but those do not work inside chrome.jar (they turn into copies).

    iconRemap: {
        // Context-sensitive actions that switch between two actions and
        // use one of those icons when customizing. This is probably ok.
        'context/blockimg': 'image/block',
        'context/frame_source': 'frame/source',
        'context/page_source': 'page/source',
        'context/page': 'menu/page',
        'context/stop': 'navigation/stop',
        'menu/context': 'menu/scroll',
        // Context menus that do not have their own submenu icon.
        // This should get fixed.
        'menu/edit': 'menu/image',
        'menu/link': 'menu/image',
        'menu/mail': 'menu/image',
        // Specialized actions that use the regular action icon.
        // This should get fixed, but I have no idea how these should look.
        'navigation/reload_skip': 'navigation/reload',
        'selection/source': 'page/source'
    },

    /**
     * Get an image by name. Names used by the menu itself are
     * menu, labels_center, labels_left_pad, labels_right_pad,
     * labels_left_edge, labels_right_edge, alt_indicator and
     * submenu_indicator_{ne,se,sw,nw}. Also called with the
     * "iconname" of all actions.
     * @returns the url to an image.
     * @type String
     */
    getImage: function(name) {
        let remappedName = this.iconRemap[name];
        if (remappedName !== undefined)
            name = remappedName;

        return this.path + name + '.png';
    }
};


/**
 * @class Controller for a displayed menu.
 *
 * @param {Theme} theme Theme instance to use.
 *   Available as this.theme after construction.
 * @param {nsIDOMElement} parentNode Node to add the menu to
 *   (normally the html "body").
 * @param {Array} layout An array of Actions.
 * @param {int} mx Initial position in pixels (client area-relative).
 * @param {int} my Initial position in pixels (client area-relative).
 * @param {bool} altPressed True to start out with the alt set of actions.
 * @param {float} textZoom Used for adjusting tooltip text size.
 * @param {bool} visible False to hide the menu initially.
 *   See {@link PieMenu#setVisible}.
 * @param context An arbitrary object stored as this.context, used by actions.
 */
function PieMenu(theme, parentNode, layout, mx, my, altPressed, textZoom,
                 visible, context) {
    this.theme = theme;
    this.x = mx;
    this.y = my;
    this.altPressed = altPressed;
    this.textZoom = textZoom;
    this.context = context;

    // index of action under mouse (or null when the mouse is in the hole)
    this._sector = null;
    // root of DOM tree containing the menu labels
    this._lnode = null;
    // label nodes
    this._normalLabels = this._boldLabels = null;

    // Create a div to contain the other nodes. This is used as just a
    // coordinate system, not a "box", with the origin at the menu's
    // center. The rest of the menu is image nodes added to this div.
    // The tooltip is inside a second div, added as a child of this
    // one in showLabels.

    // HACK: _createNode grabs the document from this.node:
    this.node = parentNode;
    this.node = this._createNode('div', 'fixed', mx + 'px', my + 'px',
                                 '0', 'auto');
    this.node.style.zIndex = '32767';

    // add the menu image
    let timg = this._createNode(
        'img', 'relative',
        '-' + this.theme.menuXOffset + 'px',
        '-' + this.theme.menuYOffset + 'px',
        'auto', 'auto');
    timg.src = this.theme.getImage('menu');
    timg.alt = '';
    this.node.appendChild(timg);

    // The icon positions are calculated once and stored here:
    this._iconX = [];
    this._iconY = [];
    // Array of action icons (no null entries):
    this._icons = [];
    // Array of "has alt action" indicators (with null entries):
    this._alticons = new Array(layout.length);
    // Array of "has submenu" indicators (with null entries):
    this._subicons = new Array(layout.length);

    for (let i = 0; i < layout.length; ++i) {
        let angle = ((i + 1) / layout.length) * 2 * Math.PI;
        this._iconX.push(-this.theme.iconXOffset + Math.round(
            this.theme.imagesRadius * Math.sin(angle)));
        this._iconY.push(-this.theme.iconYOffset - Math.round(
            this.theme.imagesRadius * Math.cos(angle)));
        this._icons.push(this._createIcon(null, i));
        // Slight hack: src is filled out by applyLayout.
    }

    this.applyLayout(layout);

    this.setVisible(visible);

    // Do this last so the menu appears in the right location.
    parentNode.appendChild(this.node);
}

PieMenu.prototype = {
    /**
     * Show or hide the entire menu.
     */
    setVisible: function(visible) {
        this.node.style.display = visible ? 'block' : 'none';
    },

    /**
     * Show a different array of Actions.
     */
    applyLayout: function(layout) {
        this.normalactions = layout.map(function(newaction, i, layout) {
            let action;
            // tweak returns null or undefined when done, so no !== here
            while (newaction != null) {
                action = newaction;
                newaction = action.tweak(this.context);
            }
            return action;
        }, this);
        this.altactions = this.normalactions.map(
            function(normalaction, i, normalactions) {
                let action = normalaction;
                let newaction = action.altaction;
                // tweak returns null or undefined when done, so no !== here
                while (newaction != null) {
                    action = newaction;
                    newaction = action.tweak(this.context);
                }
                // show/hide the alt indicator if necessary.
                // Initialized to undefined, not null, so "=="/"!=".
                if (normalaction === action) {
                    if (this._alticons[i] != null) {
                        this.node.removeChild(this._alticons[i]);
                        this._alticons[i] = null;
                    }
                } else if (this._alticons[i] == null)
                    this._alticons[i] = this._createIcon('alt_indicator', i);
                return action;
            }, this);

        this.updateEntries(this.altPressed);
    },

    /**
     * Changes the entries according to status of alt modifier.
     *
     * Also called internally by applyLayout.
     */
    updateEntries: function(altPressed) {
        this.altPressed = altPressed;
        if (altPressed)
            this.actions = this.altactions;
        else
            this.actions = this.normalactions;

        this.actions.forEach(function(action, i, actions) {
            this._icons[i].src = this.theme.getImage(action.iconname);

            // Show/hide submenu indicators if necessary:
            if (action instanceof Submenu) {
                // Initialized to undefined, not null, so "=="
                if (this._subicons[i] == null) {
                    let name = [
                        'submenu_indicator_ne', null,
                        'submenu_indicator_se', null,
                        'submenu_indicator_sw', null,
                        'submenu_indicator_nw', null,
                    ][actions.length > 4 ? i : i * 2];
                    if (name !== null)
                        this._subicons[i] = this._createIcon(name, i);
                }
            } else if (this._subicons[i] != null) {
                this.node.removeChild(this._subicons[i]);
                this._subicons[i] = null;
            }

            // Update labels if we are displaying them:
            if (this._lnode !== null) {
                this._normalLabels[i].textContent = action.label;
                this._boldLabels[i].textContent = action.label;
            }
        }, this);
    },

    /**
     * Internal helper for creating icon nodes.
     */
    _createIcon: function(path, index) {
        let timg = this._createNode(
            'img', 'absolute',
            this._iconX[index] + 'px', this._iconY[index] + 'px',
            'auto', 'auto');
        if (path !== null)
            timg.src = this.theme.getImage(path);
        timg.alt = '';
        this.node.appendChild(timg);
        return timg;
    },

    /**
     * Remove the menu.
     *
     * If doAction is true also execute the currently selected action.
     *
     * Calling any other function on the menu after this might break.
     */
    close: function(doAction) {
        // We call doAction last because some actions ("save as"
        // specifically) apparently block, and it looks weird if they
        // block with the menu open.
        this.node.parentNode.removeChild(this.node);
        if (doAction && this._sector !== null) {
            let action = this.actions[this._sector];
            // "Activates" dummies and submenus too,
            // but that's a noop so it's ok.
            if (action.active)
                action.doAction(this.context);
        }
    },

    /**
     * Update mouse position.
     *
     * @returns true if the menu moved (you may want to call hideLabels()).
     * @type bool
     */
    update: function(mx, my) {
        // identify current sector
        let nsec = null;
        let dx = mx - this.x;
        let dy = my - this.y;
        let dst2 = dx * dx + dy * dy;
        let moved = false;
        if (dst2 > this.theme.holeRadius * this.theme.holeRadius) {
            // Figure out which half-sector we are in.
            // the - 1 makes the first half of actions[0] halfsector 0.
            let hsec = Math.floor(
                this.actions.length * Math.atan2(dx, -dy) / Math.PI) - 1;
            hsec %= this.actions.length * 2;
            if (hsec < 0)
                hsec += this.actions.length * 2;
            // Slightly silly way to do integer division by 2
            // (silly js does not have an integer division operator).
            nsec = hsec >> 1;
            // If the target is a dummy adjust to the nearest neighbour.

            // We *only* do this for actual dummies, not for
            // inactive actions, or the "forward" nav action being
            // inactive makes the tab and page submenus too "sensitive".
            if (this.actions[nsec] instanceof Dummy) {
                // First or second half of the dummy?
                if (hsec % 2)
                    // Second half.
                    ++nsec;
                else
                    // First half. Make sure nsec stays positive.
                    nsec += this.actions.length - 1;
                nsec %= this.actions.length;
            }

            if (dst2 > this.theme.actionRadius * this.theme.actionRadius) {
                let action = this.actions[nsec];
                if (action instanceof Submenu) {
                    // Show submenu.

                    // The order here matters: ContextMenu changes
                    // this.context here, and that may affect tweak functions
                    // triggered by applyLayout.
                    action.activateMenu(this);
                    this.applyLayout(action.layout);
                    nsec = null;
                    this.x = mx;
                    this.y = my;
                } else {
                    moved = true;
                    let fak = this.theme.actionRadius / Math.sqrt(dst2);
                    this.x = mx - Math.round(dx * fak);
                    this.y = my - Math.round(dy * fak);
                }
                this.node.style.left = this.x + 'px';
                this.node.style.top = this.y + 'px';
            }
        }

        // Update current sector. Keep the labels visibility code
        // synced with showLabels.
        if (nsec !== this._sector) {
            // Deactivate previous rollover if there was one:
            if (this._sector !== null) {
                this._icons[this._sector].style.opacity = '1.0';
                if (this._lnode !== null) {
                    this._boldLabels[this._sector].style.visibility =
                        'hidden';
                    this._normalLabels[this._sector].style.visibility =
                        'visible';
                }
            }
            // Activate new rollover if there is one:
            if (nsec !== null && this.actions[nsec].active) {
                this._icons[nsec].style.opacity = '0.64';
                if (this._lnode !== null) {
                    this._normalLabels[nsec].style.visibility = 'hidden';
                    this._boldLabels[nsec].style.visibility = 'visible';
                }
            }
            this._sector = nsec;
        }
        return moved;
    },

    /**
     * Show the tooltip with labels.
     *
     * Calling this if it is already displayed is safe.
     *
     * The tooltip stays until hideLabels() or close() is called.
     */
    showLabels: function() {
        // create resources if necessary
        if (this._lnode === null) {
            // Main container floating behind the menu, with the
            // origin in the top left of the center image. Size is
            // picked up from the image: height so we can size the
            // label containers based on it, width so we can position
            // the righthand labels based on it.
            let main = this._createNode(
                'div', 'absolute',
                '-' + this.theme.labelsMainXOffset + 'px',
                '-' + this.theme.labelsMainYOffset + 'px',
                'auto', 'auto');
            main.style.zIndex = '-1';

            // Center image.
            let timg = this._createNode(
                'img', 'static', '0', '0', 'auto', 'auto');
            timg.src = this.theme.getImage('labels_center');
            timg.alt = '';
            main.appendChild(timg);

            // If we simply change the font weight of the labels
            // dynamically we trigger a reflow of the tooltip which
            // looks a bit ugly. So instead we use the bold labels to
            // size the background image, add the normal labels to a
            // second (transparent) div on top of that, then toggle
            // the visibility of the labels.
            //
            // Notice we still resize the label when alt is
            // pressed/released. We could avoid that too by adding
            // some more hidden labels but that is probably a bad idea
            // since their sizes may differ a lot ("Reload" vs "Reload
            // - Skip Cache"...), so a resize is better than a label
            // with a silly size.

            // Containers for the labels. Positioned next to the main
            // div. Height forced to the height of the parent to make
            // sure the entire background image shows up. Width
            // determined by the actual labels inside.
            let leftBold = this._createNode(
                'div', 'absolute', 'auto', '0', 'auto', '100%');
            leftBold.style.right = '100%';
            let leftNormal = this._createNode(
                'div', 'absolute', '0', '0', 'auto', '100%');
            leftBold.style.backgroundImage = (
                'url("' + this.theme.getImage('labels_left_pad') + '")');
            let rightBold = this._createNode(
                'div', 'absolute', '100%', '0', 'auto', '100%');
            let rightNormal = this._createNode(
                'div', 'absolute', '0', '0', 'auto', '100%');
            rightBold.style.backgroundImage = (
                'url("' + this.theme.getImage('labels_right_pad') + '")');
            // Make sure background css props are sane for both divs.
            for each (let node in [leftBold, rightBold]) {
                node.style.backgroundAttachment = 'scroll';
                node.style.backgroundPosition = 'top left';
                node.style.backgroundRepeat = 'repeat-x';
            }
            main.appendChild(leftBold);
            main.appendChild(rightBold);
            leftBold.appendChild(leftNormal);
            rightBold.appendChild(rightNormal);

            // Attach the left and right edge images to the left and
            // right divs the same way those are attached to the main div:
            let node = this._createNode(
                'img', 'absolute', 'auto', '0', 'auto', 'auto');
            node.style.right = '100%';
            node.src = this.theme.getImage('labels_left_edge');
            leftBold.appendChild(node);
            node = this._createNode('img', 'absolute', '100%', '0',
                                    'auto', 'auto');
            node.src = this.theme.getImage('labels_right_edge');
            rightBold.appendChild(node);

            // The divs containing the actual label text use relative
            // positioning to put them on the right height. Each label
            // is forced to stay on one line, so their width
            // determines the width of their containing div (which
            // makes the background the right size), and the labels
            // below show up at the right height.

            // ">> 1" is a silly way to do integer division by 2
            let rightActions = this.actions.slice(0,
                                                  this.actions.length >> 1);
            let boldLabels = rightActions.map(function(action, i, actions) {
                    return this._createLabel(this.theme.labelsRightYOffset,
                                             action.label, rightBold, true);
                }, this);
            let normalLabels = rightActions.map(function(action, i, actions) {
                    return this._createLabel(this.theme.labelsRightYOffset,
                                             action.label, rightNormal,
                                             false);
                }, this);

            // We need to add those to the container reversed.
            let leftActions = this.actions.slice(this.actions.length >> 1);
            leftActions.reverse();
            let leftLabels = leftActions.map(function(action, i, actions) {
                    return this._createLabel(this.theme.labelsLeftYOffset,
                                             action.label, leftBold, true);
                }, this);
            leftLabels.reverse();
            this._boldLabels = boldLabels.concat(leftLabels);

            leftLabels = leftActions.map(function(action, i, actions) {
                    return this._createLabel(this.theme.labelsLeftYOffset,
                                             action.label, leftNormal, false);
                }, this);
            leftLabels.reverse();
            this._normalLabels = normalLabels.concat(leftLabels);

            // hide/unhide labels (keep synced with update):
            for (let i = 0; i < this.actions.length; ++i) {
                let active = (i === this._sector &&
                              this.actions[this._sector].active);
                this._boldLabels[i].style.visibility = (
                    active ? 'visible' : 'hidden');
                this._normalLabels[i].style.visibility = (
                    active ? 'hidden' : 'visible');
            }

            // Do this last so it appears in the right spot.
            this.node.appendChild(main);
            this._lnode = main;
        }

        this._lnode.style.display = 'block';
    },

    /**
     * Helper used by showLabels.
     */
    _createLabel: function(ypos, label, container, bold) {
        let tdiv = this._createNode(
            'div', 'relative', '0', ypos + 'px',
            'auto', this.theme.labelHeight + 'px');

        // Same font value used for actual popups (see popup.css).
        tdiv.style.font = 'message-box';

        // Mozilla intentionally applies text zoom to *all* text,
        // even text with a size specified in pixels (bug 131236).
        // (IMHO that decision should be revisited now that there
        // is page zoom, but I don't know if/when that will happen).
        // Since I think in this case scaling the font up makes things
        // unreadable anyway I adjust the font size for text zoom here.
        // A better solution would be to make the tooltip graphics
        // (and possibly the menu too) scale up along with the text,
        // but I have no idea how to implement that.

        // (Changing the zoom level while the menu is displayed
        // does not work properly but I think that is acceptable).

        tdiv.style.lineHeight = Math.round(this.theme.labelHeight /
                                           this.textZoom) + 'px';
        tdiv.style.fontSize = Math.round(this.theme.fontSize /
                                         this.textZoom) + 'px';

        tdiv.style.fontWeight = bold ? 'bold' : 'normal';

        // _createNode-style cleanup for text.
        tdiv.style.letterSpacing = 'normal';
        tdiv.style.verticalAlign = 'middle';
        tdiv.style.textAlign = 'left';
        // textDecoration and textIndent are set in _createNode.
        tdiv.style.textTransform = 'none';
        // Important! without this moz tries to make every
        // label as narrow as possible.
        tdiv.style.whiteSpace = 'nowrap';
        tdiv.style.wordSpacing = 'normal';

        tdiv.textContent = label;

        container.appendChild(tdiv);
        return tdiv;
    },

    /**
     * Hide the tooltip with labels if currently visible.
     *
     * Calling this while the labels are already hidden is safe.
     */
    hideLabels: function() {
        if (this._lnode !== null)
            this._lnode.style.display = 'none';
    },

    /**
     * Internal helper undoing most style rules on a new element.
     *
     * If the element will contain text this function is not sufficient.
     *
     * @param {String} nodeName The new element's type.
     * @param {String} position CSS position value (absolute/relative/...)
     * @param {String} left CSS left position value.
     * @param {String} top CSS top position value.
     * @param {String} width CSS width value.
     * @param {String} height CSS height value.
     */
    _createNode: function(nodeName, position, left, top, width, height) {
        let node = this.node.ownerDocument.createElementNS(
            'http://www.w3.org/1999/xhtml', nodeName);
        node.style.position = position;
        node.style.left = left;
        node.style.top = top;
        node.style.width = width;
        node.style.height = height;

        // Sorted based on nsIDOMCSS2Properties.idl. All these are
        // here because a test page setting them on all div and img
        // elements affected the menu, so do not remove stuff here
        // without testing the menu (including tooltip).
        node.style.backgroundColor = 'transparent';
        node.style.backgroundImage = 'none';
        node.style.border = 'none';
        node.style.bottom = 'auto';
        node.style.clip = 'auto';
        node.style.color = 'black';
        // XXX div:before { content: 'blah' } breaks stuff.
        node.style.cursor = 'auto';
        node.style.direction = 'ltr';
        node.style.display = 'block';
        node.style.cssFloat = 'none';
        node.style.margin = '0';
        node.style.maxHeight = 'none';
        node.style.maxWidth = 'none';
        node.style.minHeight = '0';
        node.style.minWidth = '0';
        node.style.outlineStyle = 'none';
        node.style.overflow = 'visible';
        node.style.padding = '0';
        node.style.right = 'auto';
        // Just setting this on nodes containing text is apparently
        // not sufficient.
        node.style.textDecoration = 'none';
        node.style.textIndent = '0';
        node.style.verticalAlign = 'baseline';
        node.style.visibility = 'visible';

        node.style.opacity = '1';

        return node;
    }
};


/**
 * @class Basic action.
 *
 * Override doAction (subclass or reassign after instantiation) to
 * make it actually do something.
 * @param {String} label
 * @param {String} iconname Identifier for an icon image (fed to the theme).
 */
function Action(label, iconname) {
    this.label = label;
    this.iconname = iconname;
    this.altaction = null;
    this.active = true;
}

Action.prototype = {
    /**
     * Called before this action is displayed.
     *
     * Use this for context-sensitive actions or menus. It can either
     * change this action's attributes or return a different action.
     *
     * Remember this still gets called for menu previews and the like
     * where doAction will never be called, so should make fewer
     * assumptions (specifically it should not depend on browser.js).
     *
     * Tweak functions are called recursively, return null or
     * undefined when done (returning "this" triggers an infinite
     * loop).
     */
    tweak: function(context) {
        return null;
    },

    /**
     * Actually execute this action.
     */
    doAction: function(context) {}
};


/**
 * @class Special kind of action for submenus.
 *
 * @param {Array} layout An array of Actions.
 */
function Submenu(label, iconsrc, layout) {
    Action.call(this, label, iconsrc);
    this.layout = layout;
}

Submenu.prototype = {
    __proto__: Action.prototype,

    /**
     * Called right before this submenu is displayed.
     *
     * Does not have a lot of users. This interface may change.
     */
    activateMenu: function(menu) {}
};


/**
 * @class A "hole" in the menu. Special: passes focus to its neighbours.
 * @param {String} label Label to use. This argument is optional.
 */
function Dummy(label) {
    if (label === undefined)
        label = '';
    Action.call(this, label, 'dummy');
    this.active = false;
}

Dummy.prototype = {
    __proto__: Action.prototype
};
